#pragma GCC optimize("O3","unroll-loops","omit-frame-pointer","inline") //Optimization flags
#pragma GCC option("arch=native","tune=native","no-zero-upper") //Enable AVX
#pragma GCC target("avx")  //Enable AVX

#include <gnuradio/gr_complex.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <volk/volk.h>

inline static float envelope(gr_complex x)
    {
        float r_abs = std::fabs(x.real());
        float i_abs = std::fabs(x.imag());

        if(r_abs > i_abs)
            return r_abs + 0.4 * i_abs;
        else
            return i_abs + 0.4 * r_abs;
    }


float randnf()
{
	return (float)rand()/(float)(RAND_MAX);
}

gr_complex *x;
gr_complex *y;
float *pwry;
float *pwr;

void benchmark_agc_gr(struct rusage *_start,
                    struct rusage *_finish,
                    unsigned long int *_num_iterations,
		    int n)
{
    unsigned long int i;
    for (i=0; i<n; i++)
        x[i] = gr_complex(randnf(), randnf());

    getrusage(RUSAGE_SELF, _start);
    for (i=0; i<*_num_iterations; i++) {
	for(int j=0;j<n;j++) {
 		y[j]=envelope(x[j]);
	}
    }
    getrusage(RUSAGE_SELF, _finish);
    *_num_iterations *= n;
}

void init_agc_kcsa(int n)
{
	unsigned int alignment = volk_get_alignment();
	x = (gr_complex*)volk_malloc(sizeof(gr_complex)*n, alignment);
	y = (gr_complex*)volk_malloc(sizeof(gr_complex)*n, alignment);
	pwry = (float*)volk_malloc(sizeof(float)*n, alignment);
	pwr = (float*)volk_malloc(sizeof(float), alignment);
    	for (int i=0; i<n; i++) {
        	x[i] = gr_complex(randnf(), randnf());
	}
}

void benchmark_agc_kcsa(struct rusage *_start,
                    struct rusage *_finish,
                    unsigned long int *_num_iterations,
		    int n)
{
    unsigned long int i;

    float gain = 1e-3;
    float d_reference = 2.0;
    float a1 = 1.0 - 0.01;
    float a2 = 1.0 + 0.01;
    getrusage(RUSAGE_SELF, _start);
    for (i=0; i<*_num_iterations; i++) {
		volk_32fc_s32fc_multiply_32fc(y, x, gain, n);
		volk_32fc_magnitude_squared_32f(pwry, y, n);
		*pwr = 0;
		volk_32f_accumulator_s32f(pwr, pwry, n);
		*pwr /= n;
		if(*pwr > d_reference) gain *= a1;
		else gain *= a2;
    }
    getrusage(RUSAGE_SELF, _finish);
//	printf("final gain = %f\n", gain);
    *_num_iterations *= n;

}


void benchmark_agc_kcsa_2(struct rusage *_start,
                    struct rusage *_finish,
                    unsigned long int *_num_iterations,
		    int n)
{
    unsigned long int i;

    float gain = 1e-3;
    float d_reference = 2.0;
    float a1 = 1.0 - 0.01;
    float a2 = 1.0 + 0.01;
    getrusage(RUSAGE_SELF, _start);
    for (i=0; i<*_num_iterations; i++) {
		*pwr = 0;
		for(int i = 0; i < n; i++)
		{
			y[i] = x[i] * gain;
			pwry[i] = real(y[i])*real(y[i]) + imag(y[i])*imag(y[i]);
		}
		for(int i = 0; i < n; i++)
		{
			*pwr += pwry[i];
		}
		*pwr /= n;
		if(*pwr > d_reference) gain *= a1;
		else gain *= a2;
    }
    getrusage(RUSAGE_SELF, _finish);
    *_num_iterations *= n;
//	printf("final gain = %f\n", gain);

}

void benchmark_agc_kcsa_3(struct rusage *_start,
                    struct rusage *_finish,
                    unsigned long int *_num_iterations,
		    int n)
{
    unsigned long int j;

    float gain = 1e-3;
    float d_reference = 2.0;
    float a1 = 1.0 - 0.01;
    float a2 = 1.0 + 0.01;
	
    getrusage(RUSAGE_SELF, _start);
    for (j=0; j<*_num_iterations; j++) {
		*pwr = 0;
		for(int i = 0; i < n; i+=8)
		{
			y[i] = x[i] * gain;
			pwry[i] = real(y[i])*real(y[i]) + imag(y[i])*imag(y[i]);
			y[i+1] = x[i+1] * gain;
			pwry[i+1] = real(y[i+1])*real(y[i+1]) + imag(y[i+1])*imag(y[i+1]);
			y[i+2] = x[i+2] * gain;
			pwry[i+2] = real(y[i+2])*real(y[i+2]) + imag(y[i+2])*imag(y[i+2]);
			y[i+3] = x[i+3] * gain;
			pwry[i+3] = real(y[i+3])*real(y[i+3]) + imag(y[i+3])*imag(y[i+3]);
			y[i+4] = x[i+4] * gain;
			pwry[i+4] = real(y[i+4])*real(y[i+4]) + imag(y[i+4])*imag(y[i+4]);
			y[i+5] = x[i+5] * gain;
			pwry[i+5] = real(y[i+5])*real(y[i+5]) + imag(y[i+5])*imag(y[i+5]);
			y[i+6] = x[i+6] * gain;
			pwry[i+6] = real(y[i+6])*real(y[i+6]) + imag(y[i+6])*imag(y[i+6]);
			y[i+7] = x[i+7] * gain;
			pwry[i+7] = real(y[i+7])*real(y[i+7]) + imag(y[i+7])*imag(y[i+7]);
		}

		for(int i = 0; i < n; i+=8)
		{
			*pwr += pwry[i]    + pwry[i+1] + pwry[i+2] + pwry[i+3] 
			      + pwry[i+4]  + pwry[i+5] + pwry[i+6] + pwry[i+7];
		}
		*pwr /= n;
		if(*pwr > d_reference) gain *= a1;
		else gain *= a2;
    }
    getrusage(RUSAGE_SELF, _finish);
    *_num_iterations *= n;
//	printf("final gain = %f\n", gain);

}


double calculate_execution_time(struct rusage _start, struct rusage _finish)
{
    return _finish.ru_utime.tv_sec - _start.ru_utime.tv_sec
        + 1e-6*(_finish.ru_utime.tv_usec - _start.ru_utime.tv_usec)
        + _finish.ru_stime.tv_sec - _start.ru_stime.tv_sec
        + 1e-6*(_finish.ru_stime.tv_usec - _start.ru_stime.tv_usec);
}

int main(int argc, char *argv[])
{
	struct rusage s, e;

	printf("name\tn\titer\ttime\trate\n");
	printf("======================================================================\n");

	int n;
	n = 1024;
	unsigned long int iter = 1048576 * 256 / n;
	init_agc_kcsa(n);
	benchmark_agc_gr(&s, &e, &iter, n);
	double t = calculate_execution_time(s, e);
	//printf("agc_gr: %lu iter in %f sec, rate = %g\n", iter, t, iter / t);		
	printf("%s\t%d\t%lu\t%f\t%g\n", "agc_gr", n, iter, t, iter / t);

	

	n = 1;
	for(int i = 0; i < 24; i++)
	{
		unsigned long int iter = 1048576 * 256 / n;
		//unsigned long int iter = 10000000;
		init_agc_kcsa(n);
		benchmark_agc_kcsa(&s, &e, &iter, n);
		t = calculate_execution_time(s, e);
		//printf("agc_kcsa volk: n = %d, %lu iter in %f sec, rate = %g\n", n, iter, t, iter / t);
		printf("%s\t%d\t%lu\t%f\t%g\n", "agc_volk", n, iter, t, iter / t);
		n *= 2;		
	}

	n = 1;
	for(int i = 0; i < 24; i++)
	{
		unsigned long int iter = 1048576 * 256 / n;
		//unsigned long int iter = 10000000;
		init_agc_kcsa(n);
		benchmark_agc_kcsa_2(&s, &e, &iter, n);
		t = calculate_execution_time(s, e);
		//printf("agc_kcsa x1 loop: n = %d, %lu iter in %f sec, rate = %g\n", n, iter, t, iter / t);
		printf("%s\t%d\t%lu\t%f\t%g\n", "agc_x1_loop", n, iter, t, iter / t);
		n *= 2;		
	}

	n = 8;
	for(int i = 0; i < 24; i++)
	{
		unsigned long int iter = 1048576 * 256 / n;
		//unsigned long int iter = 10000000;
		init_agc_kcsa(n);
		benchmark_agc_kcsa_3(&s, &e, &iter, n);
		t = calculate_execution_time(s, e);
		//printf("agc_kcsa x8 loop: n = %d, %lu iter in %f sec, rate = %g\n", n, iter, t, iter / t);
		printf("%s\t%d\t%lu\t%f\t%g\n", "agc_x8_loop", n, iter, t, iter / t);
		n *= 2;		
	}
	return 0;
}
